////////////////////////////////////////////////////////////////////////////////
//
//  ****************************************************************************
//  * Unit Name : CreateFile_Emulation
//  * Purpose   :    CreateFile()
//  * Author    :  (Rouse_) 
//  * Copyright :  Fangorn Wizards Lab 1998 - 2012.
//  * Version   : 1.0
//  * Home Page : http://rouse.drkb.ru
//  * Home Blog : http://alexander-bagel.blogspot.com/
//  ****************************************************************************
//

program CreateFile_Emulation;

{$APPTYPE CONSOLE}

{$DEFINE DISABLE_HIDEFROMDEBUGGER}

uses
  Windows,
  SysUtils;

type
  //  STD 
  TSDTIndex = (
    sdtNtSetInformationThread,
    sdtZwOpenFile,
    sdtNtQueryObject,
    WOW64ReservedAddr);

//  64      VM 
//       32  ,   Vista  
//  4   -      
// ..   
// (    -   )
var
  SAVED_EBP: DWORD = 0;
  SAVED_ESP: DWORD = 0;

  function OpenThread(dwDesiredAccess: DWORD;
    bInheritHandle: BOOL; dwThreadId: DWORD): THandle; stdcall;
    external kernel32;

type
  NTSTATUS = DWORD;
  ULONG_PTR = DWORD;

  UNICODE_STRING = record
    Length: WORD;
    MaximumLength: WORD;
    Buffer: PWideChar;
  end;
  PUNICODE_STRING = ^UNICODE_STRING;

  TUNICODE_STRING = packed record
    Length : WORD;
    MaximumLength : WORD;
    Buffer : array [0..MAX_PATH - 1] of WideChar;
  end;

  POBJECT_ATTRIBUTES = ^OBJECT_ATTRIBUTES;
  {$EXTERNALSYM POBJECT_ATTRIBUTES}
  _OBJECT_ATTRIBUTES = record
    Length: ULONG;
    RootDirectory: THandle;
    ObjectName: PUNICODE_STRING;
    Attributes: ULONG;
    SecurityDescriptor: Pointer;
    SecurityQualityOfService: Pointer;
  end;
  {$EXTERNALSYM _OBJECT_ATTRIBUTES}
  OBJECT_ATTRIBUTES = _OBJECT_ATTRIBUTES;
  {$EXTERNALSYM OBJECT_ATTRIBUTES}
  TObjectAttributes = OBJECT_ATTRIBUTES;
  PObjectAttributes = POBJECT_ATTRIBUTES;

  _IO_STATUS_BLOCK = record
    //union {
    Status: NTSTATUS;
    //    PVOID Pointer;
    //}
    Information: ULONG_PTR;
  end;
  IO_STATUS_BLOCK = _IO_STATUS_BLOCK;
  PIO_STATUS_BLOCK = ^IO_STATUS_BLOCK;
  TIoStatusBlock = IO_STATUS_BLOCK;
  PIoStatusBlock = ^TIoStatusBlock;

  PFILE_NAME_INFORMATION = ^FILE_NAME_INFORMATION;
  FILE_NAME_INFORMATION = packed record
    FileNameLength: ULONG;
    FileName: array [0..MAX_PATH - 1] of WideChar;
  end;

  POBJECT_NAME_INFORMATION = ^TOBJECT_NAME_INFORMATION;
  TOBJECT_NAME_INFORMATION = packed record
    Name : TUNICODE_STRING;
  end;

var
  FunctionSDTIndex: array [TSDTIndex] of DWORD = (0, 0, 0, 0);

procedure _Write(const Value: AnsiString); overload;
var
  s: AnsiString;
begin
  SetLength(s, Length(Value));
  AnsiToOem(@Value[1], @s[1]);
  Writeln(s);
end;

procedure _Write(const Value: WideString); overload;
begin
  _Write(AnsiString(Value));
end;

procedure InitSDTTable;
const
  //  ,     
  ApiNames: array [TSDTIndex] of string =
    (
      'NtSetInformationThread',
      'ZwOpenFile',
      'NtQueryObject',
      ''
    );

const
  KSEG0_BASE = $80000000;
  MM_HIGHEST_USER_ADDRESS = $7FFEFFFF;
  MM_USER_PROBE_ADDRESS = $7FFF0000;
  MM_SYSTEM_RANGE_START = KSEG0_BASE;
  MustWrite = PAGE_READWRITE or PAGE_WRITECOPY or
    PAGE_EXECUTE_READWRITE or PAGE_EXECUTE_WRITECOPY;
  OBJ_CASE_INSENSITIVE = $00000040;
  FILE_SYNCHRONOUS_IO_NONALERT = $00000020;
  FILE_READ_DATA = 1;
var
  pSectionAddr, dwLength: DWORD;
  lpBuffer: TMemoryBasicInformation;
  pNtHeaders: PImageNtHeaders;
  ExportAddr: TImageDataDirectory;
  ProcessExport: Boolean;
  ImageBase: DWORD;
  IED: PImageExportDirectory;
  I: Integer;
  FuntionAddr: Pointer;
  NamesCursor: PDWORD;
  OrdinalCursor: PWORD;
  Ordinal: DWORD;
  CurrentFuncName: string;
  SDT: TSDTIndex;
begin
  //    ,    NTDLL
  pSectionAddr := GetModuleHandle('ntdll.dll');
  ImageBase := 0;
  ExportAddr.VirtualAddress := 0;
  ExportAddr.Size := 0;
  dwLength := SizeOf(TMemoryBasicInformation);

  //  WOW ,     ,
  //      sysenter
  //  32-    
  asm
    push eax
    mov  eax, fs:[$c0]
    mov  I, eax
    pop  eax
  end;

  FunctionSDTIndex[WOW64ReservedAddr] := I;

  _Write(Format('WOW64Reserved: %d', [FunctionSDTIndex[WOW64ReservedAddr]]));

  //     
  while pSectionAddr < MM_USER_PROBE_ADDRESS do
  begin

    //    
    if VirtualQuery(Pointer(pSectionAddr), lpBuffer, dwLength) <> dwLength then
      RaiseLastOSError;

    try
      //     -  
      if (lpBuffer.State = MEM_FREE) or (lpBuffer.State = MEM_RESERVE) then
        Continue;

      //    -  
      if (lpBuffer.Protect and PAGE_GUARD) = PAGE_GUARD then
        Continue;
      if (lpBuffer.Protect and PAGE_NOACCESS) = PAGE_NOACCESS then
        Continue;

      _Write(Format(' : %x', [pSectionAddr]));

      //  -       ?
      if PWord(lpBuffer.BaseAddress)^ = IMAGE_DOS_SIGNATURE then
      begin
        //    
        pNtHeaders := Pointer(Integer(lpBuffer.BaseAddress) +
          PImageDosHeader(lpBuffer.BaseAddress)^._lfanew);
        ExportAddr.VirtualAddress := 0;
        ExportAddr.Size := 0;
        ImageBase := DWORD(lpBuffer.BaseAddress);
        if (pNtHeaders^.Signature = IMAGE_NT_SIGNATURE) and
          (pNtHeaders^.FileHeader.Machine = IMAGE_FILE_MACHINE_I386) then
        begin

          _Write(' PE .');

          //   -     
          ExportAddr := pNtHeaders.OptionalHeader.DataDirectory[
            IMAGE_DIRECTORY_ENTRY_EXPORT];
          if ExportAddr.VirtualAddress <> 0 then
            Inc(ExportAddr.VirtualAddress, ImageBase)
          else
            ExportAddr.Size := 0;
        end;

        _Write(Format('  : %x', [ExportAddr.VirtualAddress]));
        _Write(Format('  : %x', [ExportAddr.Size]));

      end;

      // ,        
      ProcessExport := False;
      if ExportAddr.Size <> 0 then
        if ExportAddr.VirtualAddress >= DWORD(lpBuffer.BaseAddress) then
          ProcessExport :=
            ExportAddr.VirtualAddress + ExportAddr.Size <
            DWORD(lpBuffer.BaseAddress) + lpBuffer.RegionSize;

      //    -  
      if ProcessExport then
      begin
        if (ImageBase = 0) or (ExportAddr.VirtualAddress = 0) then Exit;
        IED := PImageExportDirectory(ExportAddr.VirtualAddress);

        _Write(Format(' : %s', [string(PAnsiChar(ImageBase + IED^.Name))]));

        // ,     ?
        if LowerCase(string(PAnsiChar(ImageBase + IED^.Name))) = 'ntdll.dll' then
        begin

          _Write('  ');

          // ,   ,     
          I := 1;
          NamesCursor := Pointer(ImageBase + DWORD(IED^.AddressOfNames));
          OrdinalCursor := Pointer(ImageBase + DWORD(IED^.AddressOfNameOrdinals));
          while I < Integer(IED^.NumberOfNames) do
          begin
            //      
            CurrentFuncName := string(PAnsiChar(ImageBase + PDWORD(NamesCursor)^));
            for SDT := sdtNtSetInformationThread to sdtNtQueryObject do
              if ApiNames[SDT] = CurrentFuncName then
              begin
                //      
                Ordinal := OrdinalCursor^ + IED^.Base;
                //      
                FuntionAddr := Pointer(ImageBase + DWORD(IED^.AddressOfFunctions));
                FuntionAddr := Pointer(ImageBase +
                  PDWORD(DWORD(FuntionAddr) + (Ordinal - 1) * 4)^);
                //      MOV
                FuntionAddr := Pointer(DWORD(FuntionAddr) + 1);
                //  SDT  
                FunctionSDTIndex[SDT] := PDWORD(FuntionAddr)^;

                _Write(Format('  %s - SDT  %d',
                  [CurrentFuncName, FunctionSDTIndex[SDT]]));

              end;
            Inc(I);
            Inc(NamesCursor);
            Inc(OrdinalCursor);
          end;
        end;
        ImageBase := 0;
      end;

      // ,     ?
      if FunctionSDTIndex[sdtNtSetInformationThread] <> 0 then
        if FunctionSDTIndex[sdtZwOpenFile] <> 0 then
          if FunctionSDTIndex[sdtNtQueryObject] <> 0 then
            Exit;
    finally
      //    ,    .
      Inc(pSectionAddr, lpBuffer.RegionSize);
    end;

  end;
end;

procedure RunDemo;
const
  THREAD_ALL_ACCESS = $000F0000 or $00100000 or $3FF;
  OBJ_CASE_INSENSITIVE = $00000040;
  FILE_SYNCHRONOUS_IO_NONALERT = $00000020;
  FILE_READ_DATA = 1;
var
  NeedInt2E: Boolean;
  SysCallArgument: DWORD;
  Status: NTSTATUS;
  UnicodeStr: UNICODE_STRING;
  oa: OBJECT_ATTRIBUTES;
  iosb: IO_STATUS_BLOCK;
  hFile: THandle;
  IsWOW64: Boolean;
  WOW64Addr: DWORD;
  dwReturn: DWORD;
  ObjectNameInfo: TOBJECT_NAME_INFORMATION;
  {$IFNDEF DISABLE_HIDEFROMDEBUGGER}
  hThread: THandle;
  {$ENDIF}
begin

  WOW64Addr := FunctionSDTIndex[WOW64ReservedAddr];
  IsWOW64 := WOW64Addr <> 0;

  // ,      Int2E
  NeedInt2E := True;
  if Win32MajorVersion > 5 then
    NeedInt2E := False
  else
    if (Win32MajorVersion = 5) and (Win32MinorVersion > 0) then
      NeedInt2E := False;

  _Write(Format('NeedInt2E: %d', [Byte(NeedInt2E)]));

  {$IFNDEF DISABLE_HIDEFROMDEBUGGER}
  //        
  // NtSetInformationThread   ThreadHideFromDebugger
  //      
  //
  // !!!
  //
  //        ,
  //    DISABLE_HIDEFROMDEBUGGER
  // ===========================================================================
  Status := $DEADBEEF;
  hThread := OpenThread(THREAD_ALL_ACCESS, False, GetCurrentThreadId);
  if LowerCase(ParamStr(1)) <> '-a' then
  begin
    _Write('  NtSetInformationThread   ThreadHideFromDebugger');
    _Write('   ');
    SysCallArgument := FunctionSDTIndex[sdtNtSetInformationThread];
    asm
      mov   SAVED_EBP, ebp
      mov   SAVED_ESP, esp

      //  

      xor  eax, eax
      push eax
      push eax
      push 17 // ThreadHideFromDebugger
      mov  eax, hThread
      push eax

      movzx eax, IsWOW64
      or   eax, eax
      jz   @32Bit

      //   64- 

      lea  eax, @64bit
      push eax
      push eax
      mov  eax, WOW64Addr
      push eax
      mov eax, SysCallArgument
      xor ecx, ecx
      lea edx, dword ptr ss:[esp+4*3]
      ret

    @64bit:
      add esp, 4
      jmp @FINALIZE

    @32Bit:

      //   32- 

      lea  eax, @FINALIZE
      push eax
      push eax
      movzx eax, NeedInt2E
      or   eax, eax
      jnz  @NT_CODE
      mov edx, esp
      mov eax, SysCallArgument
      sysenter

    @NT_CODE:

      //   W2K  

      pop eax
      lea edx, esp + 4
      mov eax, SysCallArgument
      int $2E
      nop

    @FINALIZE:
      mov Status, eax
      mov   ebp, SAVED_EBP
      mov   esp, SAVED_ESP
    end;
    _Write(Format('  %x', [Status]));
    CloseHandle(hThread);
  end;
  {$ELSE}
  _Write(' NtSetInformationThread .');
  {$ENDIF}

  //     
  // ZwOpenFile
  // ===========================================================================
  _Write('  ZwOpenFile');
  _Write('    ');
  SysCallArgument := FunctionSDTIndex[sdtZwOpenFile];

  oa.Length := SizeOf(TObjectAttributes);
  oa.RootDirectory := 0;
  oa.ObjectName := @UnicodeStr;
  oa.Attributes := OBJ_CASE_INSENSITIVE;
  oa.SecurityDescriptor := nil;
  oa.SecurityQualityOfService := nil;

  UnicodeStr.Buffer := StringToOleStr('\??\' + ParamStr(0));
  UnicodeStr.Length := Length(UnicodeStr.Buffer) * SizeOf(WideChar);
  UnicodeStr.MaximumLength := UnicodeStr.Length + SizeOf(WideChar);
  asm
    mov   SAVED_EBP, ebp
    mov   SAVED_ESP, esp

    //  

    push FILE_SYNCHRONOUS_IO_NONALERT
    push FILE_SHARE_READ + FILE_SHARE_WRITE + FILE_SHARE_DELETE
    lea  eax, iosb
    push eax
    lea  eax, oa
    push eax
    push FILE_READ_DATA + SYNCHRONIZE
    lea  eax, hFile
    push eax

    movzx eax, IsWOW64
    or   eax, eax
    jz   @32Bit

    //   64- 

    lea  eax, @64bit
    push eax
    push eax
    mov  eax, WOW64Addr
    push eax
    mov eax, SysCallArgument
    xor ecx, ecx
    lea edx, dword ptr ss:[esp+4*3]
    ret

  @64bit:
    add esp, 4
    jmp @FINALIZE

  @32Bit:

    //   32- 

    lea  eax, @FINALIZE
    push eax
    push eax
    movzx eax, NeedInt2E
    or   eax, eax
    jnz  @NT_CODE
    mov edx, esp
    mov eax, SysCallArgument
    sysenter

  @NT_CODE:

    //   W2K  

    pop eax
    lea edx, esp + 4
    mov eax, SysCallArgument
    int $2E
    nop

  @FINALIZE:
    mov Status, eax
    mov   ebp, SAVED_EBP
    mov   esp, SAVED_ESP
  end;
  if Status <> 0 then
    hFile := 0;
  _Write(Format('  %x', [Status]));
  _Write(Format(' %d', [hFile]));

  //      
  // NtQueryObject
  // ===========================================================================
  _Write('  NtQueryObject');
  _Write('     ');
  SysCallArgument := FunctionSDTIndex[sdtNtQueryObject];
  asm
    mov   SAVED_EBP, ebp
    mov   SAVED_ESP, esp

    //  

    lea   eax, dwReturn
    push  eax
    push  MAX_PATH * 2
    lea   eax, ObjectNameInfo
    push  eax
    push  1
    mov   eax, hFile
    push  eax

    movzx eax, IsWOW64
    or   eax, eax
    jz   @32Bit

    //   64- 

    lea  eax, @64bit
    push eax
    push eax
    mov  eax, WOW64Addr
    push eax
    mov eax, SysCallArgument
    xor ecx, ecx
    lea edx, dword ptr ss:[esp+4*3]
    ret

  @64bit:
    add esp, 4
    jmp @FINALIZE

  @32Bit:

    //   32- 

    lea   eax, @FINALIZE
    push  eax
    push  eax
    movzx eax, NeedInt2E
    or    eax, eax
    jnz   @NT_CODE
    mov   edx, esp
    mov   eax, SysCallArgument
    sysenter

  @NT_CODE:

    //   W2K  

    pop   eax
    lea   edx, esp + 4
    mov   eax, SysCallArgument
    int   $2E
    nop

  @FINALIZE:
    mov   Status, eax
    mov   ebp, SAVED_EBP
    mov   esp, SAVED_ESP
  end;
  _Write(Format('  %x', [Status]));
  _Write(Format('  : %s',
    [PWideChar(@ObjectNameInfo.Name.Buffer[2])]));

  if hFile <> 0 then
    CloseHandle(hFile);
end;

begin
  try
    _Write(' SDT ');
    Writeln;
    InitSDTTable;
    Writeln;
    _Write(' SDT  ');
    Writeln;
    _Write('   ');
    RunDemo;
    Writeln;
    _Write('  ');
  except
    on E: Exception do
      Writeln(E.ClassName, ': ', E.Message);
  end;
  Readln;
end.

